#!/usr/bin/env python
"""
To get quick feedback this implements an automatically reloading web server
to be run locally.

Reloading functionality is taken from the fabulous
`Werkzeug WSGI toolkit <http://www.pocoo.org/projects/werkzeug/#werkzeug>`
"""
import logging
import mimetypes
import os
import subprocess
import sys
import time
import threading
from wsgiref import simple_server, util


log = logging.getLogger(__name__)


RUN_MAIN_ENV_KEY = 'RUN_MAIN_ENV_KEY'


# todo refactor into generalized class
# todo use local.path object throughout


def restart_with_reloader():
    """Spawn a new Python interpreter with the same arguments as this one,
    but running the reloader thread.
    """
    while True:
        log.info("***restarting with reloader***")
        args = [sys.executable] + sys.argv
        new_environ = os.environ.copy()
        new_environ[RUN_MAIN_ENV_KEY] = 'true'
        exit_code = subprocess.call(args, env=new_environ)
        if exit_code != 3:
            return exit_code


def run_with_reloader(main_func, pathToWatch, pathToIgnore,
                      cleanFileNames, cleanPaths):
    """Run the given function in an independent python interpreter."""
    import signal

    signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
    if os.environ.get(RUN_MAIN_ENV_KEY) == 'true':
        t = threading.Thread(target=main_func)
        t.start()
        try:
            reloader_loop(pathToWatch, pathToIgnore,
                          cleanFileNames, cleanPaths)
        except KeyboardInterrupt:
            return

    try:
        sys.exit(restart_with_reloader())
    except KeyboardInterrupt:
        pass


def reloader_loop(pathToWatch, pathToIgnore, cleanFileNames, cleanPaths,
                  interval=0.5):
    """When this function is run from the main thread, it will force other
    threads to exit when any files passed in here change..

    Copyright notice: this function is based on ``_reloader_stat_loop()``
    from Werkzeug which is based on autoreload.py
    from CherryPy trac which originated from WSGIKit which is now dead.

    :param LocalPath pathToWatch: path of the directory to be watched.
    """
    pathTimeMap = {}
    while True:
        # for p in cleanPaths:
        #     if not p.exists():
        #         log.info("%s does not exist: trigger build..." % (p))
        #         sys.exit(3)

        paths = [p for p in pathToWatch.walk(
            filter=lambda p: "\\." not in p._path and "/." not in p._path)
            if not p._path.startswith(pathToIgnore._path) and
            "\\." not in p._path and "/." not in p._path and not
            p.isdir()]
        changedPaths = []
        for filePath in paths:
            try:
                mtime = filePath.stat().st_mtime
            except OSError:
                log.warning("problem with %s" % (filePath), exc_info=True)
                continue

            oldTime = pathTimeMap.get(filePath)
            if oldTime is None:
                pathTimeMap[filePath] = mtime
                continue

            elif mtime > oldTime:
                changedPaths.append(filePath.basename)
                if cleanFileNames == "ALL":
                    break

        if changedPaths:
            # fixme handle changes in build dir intelligently (query git)
            log.info("detected changes in %s: reloading" % (changedPaths))
            if (cleanFileNames == "ALL" or
                    any(n in [b.basename for b in changedPaths]
                        for n in cleanFileNames)):
                log.info("cleaning necessary in %s" % (cleanPaths))
                for path in cleanPaths:
                    log.info("cleaning %s" % (path))
                    path.delete()
            log.info("trigger build ...")
            sys.exit(3)

        time.sleep(interval)


def make_server(path, port):
    def minimal_wsgi_app(environ, respond):
        """simple wsgi app to serve html files"""
        fn = os.path.join(path, environ['PATH_INFO'][1:])
        if '.' not in fn.split(os.path.sep)[-1]:
            fn = os.path.join(fn, 'index.html')
        type_ = mimetypes.guess_type(fn)[0]
        if os.path.exists(fn):
            respond('200 OK', [('Content-Type', type_)])
            return util.FileWrapper(open(fn, "rb"))

        respond('404 Not Found', [('Content-Type', 'text/plain')])
        return ['404 not found']

    return simple_server.make_server('', port, minimal_wsgi_app)


def serve_with_reloader(
        serveFromPath, port, changedCallback, pathToWatch, pathToIgnore,
        cleanFileNames=None, cleanPaths=None):
    """

    :param serveFromPath: path to the folder to be served
    :param pathToWatch: path to watch recursively for changed files
    :param port: port to be served ond
    :param changedCallback: function to be called if a monitored file changes
    :param pathToIgnore: don't watch this path for changes
    :param cleanFileNames: list of filenames that should trigger a full clean
    :param cleanPaths: paths to delete on a full clean
    """
    def call_func_then_serve():
        """Calls the passed in function and then starts the server"""
        result = changedCallback()
        log.info("call %s -> %s" % (changedCallback.__name__, result))
        server = make_server(serveFromPath._path, port)
        log.info("serve %s on http://localhost:%s, control-C to stop" %
                 (serveFromPath, port))
        server.serve_forever()

    log.info("Serve while watching folder %s" % (pathToWatch))
    run_with_reloader(call_func_then_serve, pathToWatch, pathToIgnore,
                      cleanFileNames or [], cleanPaths or [])



#!/usr/bin/env python
from __future__ import print_function

import logging
import sys
import imp

from plumbum import cli, cmd, local
from plumbum.commands import ProcessExecutionError
import plumbum.path.utils as plumbum_utils

from loslassa import __version__

from loslassa.devserver import serve_with_reloader
from loslassa.utils import (
    simple_dbg, find_file, adjust_log,
    friendly_exception_handler, LoslassaError)


LOSLASSA = "loslassa"


log = logging.getLogger()


class LoslassaProject(object):
    SPHINX_CONFIG = "conf.py"
    HERE = local.path(__file__).dirname
    PROJECTS = HERE/"projects"
    EXAMPLE_PROJECT = PROJECTS/"example"
    SKELETON_PROJECT = PROJECTS/"skeleton"

    def __init__(self, projectPath):
        assert projectPath, "No project path set"
        self.projectPath = projectPath
        self.inputContainer = local.path(local.cwd/projectPath)
        self.projectName = self.inputContainer.basename
        self.sphinxConfig = self.inputContainer/self.SPHINX_CONFIG
        self.buildPath = local.path(self.inputContainer/"__build")
        self.doctreesPath = self.buildPath/"doctrees"
        self.outputPath = self.buildPath/"html"
        log.info("[PROJECT INFO]: input from %s - html generated in %s" %
                 (self.inputContainer, self.outputPath))

    def __str__(self):
        return simple_dbg(self)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.inputContainer)

    def create_project(self):
        plumbum_utils.copy(self.SKELETON_PROJECT, self.inputContainer)
        log.info("created project at %s" % (self.inputContainer))

    @property
    def isProject(self):
        return self.sphinxConfig.exists()

    def generate_html(self):
        args = ["-v", "-b", "dirhtml", "-d", self.doctreesPath._path,
                self.inputContainer._path, self.outputPath._path]
        log.debug("sphinx-build %s" % (" ".join(args)))
        output = cmd.sphinx_build(*args)
        log.info("sphinx output...\n%s" % (output))

    def add_new_files(self):
        # fixme use GitPorcelainPorcelain
        with local.cwd(self.projectPath):
            log.info(cmd.git("add", "--all", "."))

    def commit_files(self):
        # fixme use GitPorcelainPorcelain
        with local.cwd(self.projectPath):
            log.info(cmd.git("commit", "-m", "new build"))

    def commit_all(self):
        self.add_new_files()
        try:
            self.commit_files()
        except ProcessExecutionError as e:
            if "nothing to commit" not in e.stdout:
                raise

    def build_generate_only(self):
        self.generate_html()

    def build_and_autocommit(self):
        self.generate_html()
        self.commit_all()


class LoslassaConfig(object):
    """Access to Loslassa settings in configuration file"""
    def __init__(self, projectPath, configName="conf"):
        configPath = local.path(projectPath, configName + ".py")
        assert configPath.exists(), configPath
        fp, path, suffixes = imp.find_module(configName, [projectPath._path])
        try:
            self.conf = imp.load_module(configName, fp, path, suffixes)
        finally:
            fp.close()
        self.settings = self.conf.LoslassaSettings

    def __str__(self):
        return simple_dbg(self)

    def __getattr__(self, item):
        return getattr(self.settings, item)


class GitPorcelainPorcelain(object):
    def __init__(self, projectPath):
        self.projectPath = projectPath
        self.projectName = projectPath.basename
        self.settings = LoslassaConfig(projectPath)
        self.gitPath = self.projectPath/".git"

    def create_repo(self):
        with local.cwd(self.projectPath):
            log.info(cmd.git("init"))
            log.info(cmd.git("add", "."))
            log.info(cmd.git("commit", "-m", "initial commit"))

    def connect_project(self):
        """
        http://mikeeverhart.net/git/using-git-to-deploy-code/

        local
        -----
        prepare local project:
            cd into projectPath
            git remote add www ssh://<sshUser>@<remoteFqdn>/<bareclonepath>

        prepare remote bare repo:
            git clone --bare <proj> <proj>.git
            cd into bare clone repo
                $ cat > hooks/post-receive
                #!/bin/sh
                GIT_WORK_TREE=<remote bare clones path> git checkout -f
            chmod +x hooks/post-receive

        to remote:
            rsync -avx
                <proj>.git bestuebe@best-uebersetzungen.de: ./projects/<proj>/

        remote
        ------
        git clone ~/projects/<proj>.git ~/www_content/<proj>
        ln -s
            ~/www_content/bilderwerkstatt_ravensburg .de/__build/html
            <dir containing web content>

        initial push
        ------------
        git push www +master:refs/heads/master

        all other pushes
        ----------------
        git push www master
        """
        # todo
        # todo look into plumbum remote path
        # ... or use posixpath
        # import posixpath as pp

        # fixme just a sketch check these paths - very likely wrong
        # self.bareCLoneName = self.projectName + ".git"
        # self.bareCLonePath = self.projectPath + ".git"
        # remHomePath = "/home/%s" % (self.sshUser)
        # remoteBareClonesContainer = pp.join(remHomePath, "projects")
        # remoteBareClonePath = pp.join(
        #     remoteBareClonesContainer, self.bareCLoneName)
        # remoteContentsContainer = pp.join(remHomePath, "www_content")
        # remoteContentsPath = pp.join(remHomePath, "www_content")

    @property
    def isRepo(self):
        return self.gitPath.exists()

    @property
    def sshOptions(self):
        return (
            ["-i", self.settings.privateKeyPath,
             "%s@%s" % (self.settings.sshUser, self.settings.remoteFqdn)])

    @property
    def sshUser(self):
        return self.settings.sshUser

    @property
    def remoteFqdn(self):
        return self.settings.remoteFqdn

    @property
    def privateKeyPath(self):
        return self.settings.privateKeyPath

    @property
    def sshIsOk(self):
        try:
            cmd.ssh(self.sshOptions + ["id"])
            return True

        except:
            log.warning("ssh connection failed", exc_info=True)
            return False


class LoslassaCliApplication(cli.Application):
    PROGNAME = LOSLASSA
    VERSION = __version__
    USAGE = LOSLASSA + " [start|play|loslassa] [OPTIONS]"
    projectPath = None
    logLevel = logging.DEBUG
    logFilePath = None

    def __str__(self):
        return simple_dbg(
            self, excludeNames=["parent", "nested_command"])

    @cli.autoswitch(str)
    def project_name(self, projectName):
        """Set name (can be a relative or absolute path as well"""
        self.projectPath = local.path(projectName)

    @cli.autoswitch(str)
    def verbosity(self, level):
        """Adjust the talkativeness of loslassing activities.

        :param str level: log level (one of the accepted logging values)
        Levels from very chatty to almost silent: debug, info, warning, error
        """
        self.logLevel = level

    @cli.autoswitch(str)
    def log_to_file(self, filePath):
        """Log to a file instead of the console"""
        self.logFilePath = filePath

    def _init(self, create=False):
        if not self.projectPath:
            log.warning("no conf.py here ... searching (press CTRL-C to stop)")
            confPath = find_file(local.cwd, LoslassaProject.SPHINX_CONFIG)
            self.projectPath = local.path(confPath.dirname)
        self.project = LoslassaProject(self.projectPath)
        if create:
            self.project.create_project()
        adjust_log(level=self.logLevel, filePath=self.logFilePath)
        log.info("working with project '%s'" % (self.project.projectName))
        if log.getEffectiveLevel() > logging.DEBUG:
            sys.excepthook = friendly_exception_handler


class Loslassa(LoslassaCliApplication):
    def main(self, *args):
        log.debug("executing command %s" % str(self.nested_command))
        if args:
            log.error("unknown command %r" % (args[0]))
            return 1

        if not self.nested_command:
            log.error("Which kind of loslassing? "
                      "Try %s --help" % (Loslassa.PROGNAME))
            return 1


@Loslassa.subcommand("start")
class LoslassaStart(LoslassaCliApplication):
    """Starts a new project by creating the initial project structure"""
    def main(self):
        log.info("start loslassing ...")
        if not self.projectPath:
            raise LoslassaError("please provide a name for the project")

        if self.projectPath.exists():
            raise LoslassaError(
                "'%s' already exists (try a different name?)." %
                (self.projectPath.basename))

        self._init(create=True)
        log.info("Created project '%s' at %s" %
                 (self.project.projectName, self.project.inputContainer))


@Loslassa.subcommand("play")
class LoslassaPlay(LoslassaCliApplication):
    """Start playing with the source and create your page"""
    serverPort = 8080
    autocommit = True

    @cli.autoswitch()
    def no_autocommit(self):
        """switch off automatic commits to the repository"""
        self.autocommit = False

    @cli.autoswitch(int)
    def serve_on_port(self, serverPort):
        """Set port manually"""
        self.serverPort = serverPort

    def _init(self, create=False):
        super(LoslassaPlay, self)._init(create)
        # fixme reloader reloads main method instead just the server!?
        if not self.project.sphinxConfig.exists():
            raise LoslassaError(
                "no config found at %s" % (self.project.sphinxConfig))

    def main(self):
        """Create the project representation and start serving"""
        log.info("play loslassing...")
        self._init()
        buildCommand = (self.project.build_and_autocommit if self.autocommit
                        else self.project.build_generate_only)
        serve_with_reloader(
            serveFromPath=self.project.outputPath,
            port=self.serverPort,
            changedCallback=buildCommand,
            pathToWatch=self.project.inputContainer,
            pathToIgnore=self.project.buildPath,
            # cleanFileNames=["conf", "index"],
            cleanFileNames="ALL",
            cleanPaths=[self.project.outputPath, self.project.doctreesPath])


@Loslassa.subcommand(LOSLASSA)
class LoslassaLoslassa(LoslassaCliApplication):
    """Practice loslassing by pushing your page into the interwebs"""
    def main(self):
        self._init(create=True)
        # todo make a progress bar consisting of loslassa :)
        log.info("loslassa loslassa loslassa ...")
        # raise LoslassaError("coming soon...")


def main():
    logging.basicConfig(level=logging.INFO, format='%(message)s')
    if len(sys.argv) == 1:
        log.info("no args ... using test config instead")
        name = "/home/obestwalter/work/linda/bilderwerkstatt_ravensburg.de"
        args = ["play", "--no-autocommit",
                "--verbosity", "DEBUG", "--project-name", name]
        sys.argv.extend(args)
    Loslassa.run()


if __name__ == "__main__":
    sys.exit(main())



print(
    " ".join(
        ['/usr/local/bin/sphinx-build', '-b', 'dirhtml', '-d',
         '/home/obestwalter/projects/linda/bilderwerkstatt_ravensburg.de/'
         '__build/doctrees',
         '/home/obestwalter/projects/linda/bilderwerkstatt_ravensburg.de',
         '/home/obestwalter/projects/linda/bilderwerkstatt_ravensburg.de/'
         '__build/html']
    ))



"""

.pypirc should look like this::

    [server-login]
    username:oliver.bestwalter
    password:<wont tell ya>

Release process:
 * update __version__
 * python ./setup.py sdist upload
"""
from setuptools import setup

from loslassa import __version__

with open("requirements.txt") as f:
    requirements = f.readlines()

setup(
    name='Loslassa',
    description='Just another toy static website generator',
    version=__version__,
    license='BSD',
    url='http://github.com/obestwalter/loslassa/',
    author='Oliver Bestwalter',
    author_email='oliver@bestwalter.de',
    long_description='',
    packages=['loslassa'],
    include_package_data=True,
    zip_safe=False,
    platforms='Unix',
    install_requires=requirements,
    entry_points=dict(console_scripts=['loslassa=loslassa._loslassa:main']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



import logging
import traceback

from plumbum.machines.local import local


__all__ = [
    "adjust_log",
    "cached_property",
    "find_file",
    "friendly_exception_handler",
    "LoslassaError",
    "simple_dbg",
    "UtilsError",
]

log = logging.getLogger()


def adjust_log(level, filePath=None):
    """Add file handler to logging. If set to debug output extra info

    :param level: log level (one of the logging level constants)
    :param filePath: path to where the logfile should be written to
    """
    fmt = '%(message)s'
    dateFmt = '%Y-%m-%d-%H:%M:%S'
    log.setLevel(level)
    if log.getEffectiveLevel() <= logging.DEBUG:
        fmt = ('%(asctime)s %(funcName)s [%(lineno)s] %(levelname)s: ' + fmt)
    formatter = logging.Formatter(fmt=fmt, datefmt=dateFmt)
    if filePath:
        log.addHandler(logging.FileHandler(filePath))
    for handler in log.handlers:
        handler.setFormatter(formatter)


def simple_dbg(obj, excludeNames=None):
    """pretty print information about an object

    use for quick overview while developing and for __str__ if suitable

    :param object obj: object to be print
    :param list excludeNames: list of attrs to exclude
    :rtype: str
    """
    excludeNames = excludeNames or []
    lines = ["%s" % repr(obj)]
    for name in sorted([n for n in dir(obj) if n not in excludeNames]):
        try:
            attr = getattr(obj, name)
            content = str(attr).replace("\n", "\n|    ")
            lines.append("%s %s: %s" % (name, type(attr), content))
        except Exception as e:
            msg = "[EXC]%s: %s: %s" % (name, e.__class__.__name__, e.message)
            lines.append(msg)
    return "\n".join(
        [line for line in lines
         if line and 'method' not in line and not line.startswith('_')])


def find_file(searchPath, wantedName):
    with local.cwd(searchPath):
        matches = [p for p in local.cwd.walk() if p.basename == wantedName]
    if len(matches) == 1:
        return matches[0]

    if not matches:
        raise UtilsError("%s not found in %s" % (wantedName, searchPath))

    raise UtilsError(
        "too many projects found in %s: %s" % (searchPath, matches))


def friendly_exception_handler(exception_type, exception, tb):
    if isinstance(exception_type, LoslassaError):
        msg = exception.message
    else:
        msg = (
            "Whoops that was uncalled for ... sorry about that :(\n"
            "An error of %s occurred.\n"
            "The error message was '%s'" % (exception_type, exception))
    rbString = ''.join(traceback.format_tb(tb))
    msg += (
        "What follows might not say much to you, but somebody who knows "
        "Python will be able to figure out what went wrong:\n%s" % (rbString))
    log.error(msg)


class cached_property(object):
    """read-only property that's only evaluated once (filched from Werkzeug)"""
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, _):
        if obj is None:
            return self

        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class LoslassaError(Exception):
    pass


class UtilsError(LoslassaError):
    pass
